import numpy as np
import nibabel as nib
from nilearn.input_data import NiftiMasker
from sklearn.utils.testing import assert_array_almost_equal

from fmralign import metrics

def test_score_voxelwise():
    A = np.asarray([[
        [[1, 1.2, 1, 1.2, 1]],
        [[1, 1, 1, .2, 1]], 
        [[1, -1, 1, -1, 1]]
    ]])
    B = np.asarray([[
        [[0, 0.2, 0, 0.2, 0]],
        [[.2, 1, 1, 1, 1]],
        [[-1, 1, -1, 1, -1]]
    ]])
    im_A = nib.Nifti1Image(A, np.eye(4))
    im_B = nib.Nifti1Image(B, np.eye(4))
    mask_img = nib.Nifti1Image(np.ones(im_A.shape[0:3]), np.eye(4))
    masker = NiftiMasker(mask_img=mask_img).fit()

    # check correlation raw_values
    correlation1 = metrics.score_voxelwise(im_A, im_B,
                                          masker, loss='corr')
    assert_array_almost_equal(correlation1, [1., -0.25, -1])

    # check correlation uniform_average
    correlation2 = metrics.score_voxelwise(im_A, im_B,
                                          masker, loss='corr',
                                          multioutput='uniform_average')
    assert(correlation2.ndim == 0)

    # check R2
    r2 = metrics.score_voxelwise(im_A, im_B, masker, loss='R2')
    assert_array_almost_equal(r2, [-1., -1., -1.])

    # check normalized reconstruction
    norm_rec = metrics.score_voxelwise(im_A, im_B, masker,
                                       loss='n_reconstruction_err')
    assert_array_almost_equal(norm_rec, [0.14966, 0.683168, -1.])


def test_normalized_reconstruction_error():
    A = np.asarray([
        [1, 1.2, 1, 1.2, 1],
        [1, 1, 1, .2, 1],
        [1, -1, 1, -1, 1]
    ])
    B = np.asarray([
        [0, 0.2, 0, 0.2, 0],
        [.2, 1, 1, 1, 1],
        [-1, 1, -1, 1, -1]
    ])
    
    avg_norm_rec = metrics.normalized_reconstruction_error(
        A, B, multioutput='uniform_average')
    np.testing.assert_almost_equal(avg_norm_rec, -0.788203)



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.utils.testing import assert_array_almost_equal, assert_greater
from scipy.linalg import orthogonal_procrustes
from fmralign.alignment_methods import scaled_procrustes, \
    optimal_permutation, _voxelwise_signal_projection
from fmralign.alignment_methods import Identity, DiagonalAlignment, Hungarian,\
    ScaledOrthogonalAlignment, RidgeAlignment, OptimalTransportAlignment
from fmralign.tests.utils import assert_class_align_better_than_identity, \
    zero_mean_coefficient_determination


def test_scaled_procrustes_algorithmic():
    '''Test Scaled procrustes'''
    X = np.random.randn(10, 20)
    Y = np.zeros_like(X)
    R = np.eye(X.shape[1])
    R_test, _ = scaled_procrustes(X, Y)
    assert_array_almost_equal(R, R_test)

    '''Test if scaled_procrustes basis is orthogonal'''
    X = np.random.rand(3, 4)
    X = X - X.mean(axis=1, keepdims=True)

    Y = np.random.rand(3, 4)
    Y = Y - Y.mean(axis=1, keepdims=True)

    R, _ = scaled_procrustes(X.T, Y.T)
    assert_array_almost_equal(R.dot(R.T), np.eye(R.shape[0]))
    assert_array_almost_equal(R.T.dot(R), np.eye(R.shape[0]))

    ''' Test if it sticks to scipy scaled procrustes in a simple case'''
    X = np.random.rand(4, 4)
    Y = np.random.rand(4, 4)

    R, _ = scaled_procrustes(X, Y)
    R_s, _ = orthogonal_procrustes(Y, X)
    assert_array_almost_equal(R.T, R_s)

    '''Test that primal and dual give same results'''
    # number of samples n , number of voxels p
    n, p = 100, 20
    X = np.random.randn(n, p)
    Y = np.random.randn(n, p)
    R1, s1 = scaled_procrustes(X, Y, scaling=True, primal=True)
    R_s, _ = orthogonal_procrustes(Y, X)
    R2, s2 = scaled_procrustes(X, Y, scaling=True, primal=False)
    assert_array_almost_equal(R1, R2)
    assert_array_almost_equal(R2, R_s.T)
    n, p = 20, 100
    X = np.random.randn(n, p)
    Y = np.random.randn(n, p)
    R1, s1 = scaled_procrustes(X, Y, scaling=True, primal=True)
    R_s, _ = orthogonal_procrustes(Y, X)
    R2, s2 = scaled_procrustes(X, Y, scaling=True, primal=False)
    assert_array_almost_equal(s1 * X.dot(R1), s2 * X.dot(R2))


def test_scaled_procrustes_on_simple_exact_cases():
    '''Orthogonal Matrix'''
    v = 10
    k = 10
    rnd_matrix = np.random.rand(v, k)
    R, _ = np.linalg.qr(rnd_matrix)
    X = np.random.rand(10, 20)
    X = X - X.mean(axis=1, keepdims=True)
    Y = R.dot(X)
    R_test, _ = scaled_procrustes(X.T, Y.T)
    assert_array_almost_equal(R_test.T, R)

    '''Scaled Matrix'''
    X = np.array([[1., 2., 3., 4.],
                  [5., 3., 4., 6.],
                  [7., 8., -5., -2.]])

    X = X - X.mean(axis=1, keepdims=True)

    Y = 2 * X
    Y = Y - Y.mean(axis=1, keepdims=True)

    assert_array_almost_equal(
        scaled_procrustes(X.T, Y.T, scaling=True)[0], np.eye(3))
    assert_array_almost_equal(scaled_procrustes(X.T, Y.T, scaling=True)[1], 2)

    '''3D Rotation'''
    R = np.array([[1., 0., 0.], [0., np.cos(1), -np.sin(1)],
                  [0., np.sin(1), np.cos(1)]])
    X = np.random.rand(3, 4)
    X = X - X.mean(axis=1, keepdims=True)
    Y = R.dot(X)

    R_test, _ = scaled_procrustes(X.T, Y.T)
    assert_array_almost_equal(
        R.dot(np.array([0., 1., 0.])),
        np.array([0., np.cos(1), np.sin(1)])
    )
    assert_array_almost_equal(
        R.dot(np.array([0., 0., 1.])),
        np.array([0., -np.sin(1), np.cos(1)])
    )
    assert_array_almost_equal(R, R_test.T)

    '''Test Scaled_Orthogonal_Alignment on an exact case'''
    ortho_al = ScaledOrthogonalAlignment(scaling=False)
    ortho_al.fit(X.T, Y.T)
    assert_array_almost_equal(
        ortho_al.transform(X.T),
        Y.T)


def test_optimal_permutation_on_translation_case():
    ''' Test optimal permutation method'''
    X = np.array([[1., 4., 10], [1.5, 5, 10], [1, 5, 11], [1, 5.5, 8]]).T
    # translate the data matrix along features axis (voxels are permutated)
    Y = np.roll(X, 2, axis=1)

    opt = optimal_permutation(X, Y).toarray()
    assert_array_almost_equal(opt.dot(X.T).T, Y)

    U = np.vstack([X.T, 2 * X.T])
    V = np.roll(U, 4, axis=1)

    opt = optimal_permutation(U, V).toarray()
    assert_array_almost_equal(opt.dot(U.T).T, V)


def test_projection_coefficients():
    n_samples = 4
    n_features = 6
    A = np.random.rand(n_samples, n_features)
    C = []
    for i, a in enumerate(A):
        C.append((i + 1) * a)
    c = _voxelwise_signal_projection(A, C, 2)
    assert_array_almost_equal(c, [i + 1 for i in range(n_samples)])


def test_all_classes_R_and_pred_shape_and_better_than_identity():
    from scipy.sparse.csc import csc_matrix
    '''Test all classes on random case'''
    # test on empty data
    X = np.zeros((30, 10))
    for algo in [Identity(), RidgeAlignment(), ScaledOrthogonalAlignment(),
                 OptimalTransportAlignment(), Hungarian(), DiagonalAlignment()]:
        algo.fit(X, X)
        assert_array_almost_equal(algo.transform(X), X)
    # if trying to learn a fit from array of zeros to zeros (empty parcel)
    # every algo will return a zero matrix

    for n_samples, n_features in [(100, 20), (20, 100)]:
        X = np.random.randn(n_samples, n_features)
        Y = np.random.randn(n_samples, n_features)
        id = Identity()
        id.fit(X, Y)
        identity_baseline_score = zero_mean_coefficient_determination(Y, X)
        assert_array_almost_equal(X, id.transform(X))
        for algo in [RidgeAlignment(), ScaledOrthogonalAlignment(),
                     ScaledOrthogonalAlignment(scaling=False),
                     OptimalTransportAlignment(),
                     Hungarian(), DiagonalAlignment()]:
            # print(algo)
            algo.fit(X, Y)
            # test that permutation matrix shape is (20, 20) except for Ridge
            if type(algo.R) == csc_matrix:
                R = algo.R.toarray()
                assert(R.shape == (n_features, n_features))
            elif type(algo) != RidgeAlignment:
                R = algo.R
                assert(R.shape == (n_features, n_features))
            # test pred shape and loss improvement compared to identity
            X_pred = algo.transform(X)
            assert(X_pred.shape == X.shape)
            algo_score = zero_mean_coefficient_determination(
                Y, X_pred)
            assert_greater(algo_score, identity_baseline_score)



# -*- coding: utf-8 -*-
import os
from nilearn.datasets.utils import _fetch_files, _get_dataset_dir
import pandas as pd


def fetch_ibc_subjects_contrasts(subjects, data_dir=None, verbose=1):
    """Fetch all IBC contrast maps for each of subjects.
    After downloading all relevant images that are not already cached,
    it returns a dataframe with all needed links.

    Parameters
    ----------
    subjects : list of str.
        Subjects data to download. Available strings are ['sub-01', 'sub-02',
        'sub-04' ... 'sub-09', 'sub-11' ... sub-15]
    data_dir: string, optional
        Path of the data directory. Used to force data storage in a specified
        location.
    verbose: int, optional
        verbosity level (0 means no message).

    Returns
    -------
    files : list of list of str
        List (for every subject) of list of path (for every conditions),
        in ap then pa acquisition.
    metadata_df : Pandas Dataframe
        Table containing some metadata for each available image in the dataset,
        as well as their pathself.
        Filtered to contain only the 'subjects' parameter metadatas
    mask: str
        Path to the mask to be used on the data
    Notes
    ------
    This function is a caller to nilearn.datasets.utils._fetch_files in order
    to simplify examples reading and understanding for fmralign.
    See Also
    ---------
    nilearn.datasets.fetch_localizer_calculation_task
    nilearn.datasets.fetch_localizer_contrasts
    """
    # The URLs can be retrieved from the nilearn account on OSF
    if subjects is "all":
        subjects = ['sub-%02d' %
                    i for i in [1, 2, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15]]
    dataset_name = 'ibc'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir,
                                verbose=verbose)

    # download or retrieve metadatas, put it in a dataframe,
    # list all condition and specify path to the right directory
    metadata_path = _fetch_files(data_dir, [('ibc_3mm_all_subjects_metadata.csv',
                                             "https://osf.io/pcvje/download",
                                             {"uncompress": True})],
                                 verbose=verbose)
    metadata_df = pd.read_csv(metadata_path[0])
    conditions = metadata_df.condition.unique()
    metadata_df['path'] = metadata_df['path'].str.replace(
        'path_to_dir', data_dir)
    # filter the dataframe to return only rows relevant for subjects argument
    metadata_df = metadata_df[metadata_df.subject.isin(subjects)]

    # download / retrieve mask niimg and find its path
    mask = _fetch_files(
        data_dir, [('gm_mask_3mm.nii.gz', "https://osf.io/yvju3/download",
                    {"uncompress": True})], verbose=verbose)[0]

    # list all url keys for downloading separetely each subject data
    url_keys = {"sub-01": "8z23h", "sub-02": "e9kbm", "sub-04": "qn5b6",
                "sub-05": "u74a3", "sub-06": "83bje", "sub-07": "43j69",
                "sub-08": "ua8qx", "sub-09": "bxwtv", "sub-11": "3dfbv",
                "sub-12": "uat7d", "sub-13": "p238h", "sub-14": "prdk4",
                "sub-15": "sw72z"}

    # for all subjects in argument, download all contrasts images and list
    # their path in the variable files
    opts = {'uncompress': True}
    files = []
    for subject in subjects:
        url = "https://osf.io/%s/download" % url_keys[subject]
        filenames = [(os.path.join(subject, "%s_ap.nii.gz" % condition),
                      url, opts) for condition in conditions]
        filenames.extend([(os.path.join(subject, "%s_pa.nii.gz" % condition),
                           url, opts) for condition in conditions])
        files.append(_fetch_files(data_dir, filenames, verbose=verbose))
    return files, metadata_df, mask



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.utils.testing import assert_array_almost_equal
from nilearn.input_data import NiftiMasker
import nibabel
import pytest
from fmralign.tests.utils import random_niimg
from fmralign._utils import _make_parcellation, _hierarchical_k_means
import nilearn
from packaging import version
from nilearn._utils.data_gen import generate_fake_fmri
from itertools import product


def test_hierarchical_k_means():
    X = [[10, -10, 30], [12, -8, 24]]
    truth_labels = np.tile([0, 1, 2], 5)
    X = np.tile(X, 5).T
    test_labels = _hierarchical_k_means(X, 3)
    truth_labels = np.tile([test_labels[0], test_labels[1], test_labels[2]], 5)
    assert_array_almost_equal(test_labels, truth_labels)


def test_make_parcellation():
    # make_parcellation is built on Nilearn which already has several test for its Parcellation class
    # here we test just the call of the API is right on a simple example
    img, mask_img = random_niimg((7, 6, 8, 5))
    indexes = np.arange(img.shape[-1])
    masker = NiftiMasker(mask_img=mask_img).fit()

    methods = ["kmeans", "ward", "hierarchical_kmeans"]

    # check rena only if nilearn version allow it
    if version.parse(nilearn.__version__) <= version.parse("0.5.2"):
        with pytest.raises(Exception):
            assert _make_parcellation(img, "rena", n_pieces, masker)
    else:
        methods.append("rena")

    for clustering_method in methods:
        # check n_pieces = 1 gives out ones of right shape
        assert (_make_parcellation(
            img, indexes, clustering_method, 1, masker) == masker.transform(mask_img)).all()

        # check n_pieces = 2 find right clustering
        labels = _make_parcellation(
            img, indexes, clustering_method, 2, masker)
        assert(len(np.unique(labels)) == 2)

        # check that not inputing n_pieces yields problems
        with pytest.raises(Exception):
            assert _make_parcellation(
                img, indexes, clustering_method, 0, masker)

    clustering = nibabel.Nifti1Image(
        np.hstack([np.ones((7, 3, 8)), 2 * np.ones((7, 3, 8))]), np.eye(4))

    # check 3D Niimg clusterings
    for n_pieces in [0, 1, 2]:
        labels = _make_parcellation(
            img, indexes, clustering, n_pieces, masker)
        assert(len(np.unique(labels)) == 2)

    # check warning if a parcel is too big
    with pytest.warns(UserWarning):
        clustering = nibabel.Nifti1Image(
            np.hstack([np.ones(2000), 4 * np.ones(800)]), np.eye(4))
        _make_parcellation(
            img, indexes, clustering_method, n_pieces, masker)



# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import pearsonr
from sklearn.metrics import r2_score


def score_voxelwise(ground_truth, prediction, masker, loss,
                    multioutput='raw_values'):
    """
    Calculates loss function for predicted, ground truth
    arrays. Supported scores are R2, correlation, and normalized
    reconstruction error (Bazeille et al., 2019)

    Parameters
    ----------
    ground_truth: 3D or 4D Niimg
        Reference image
    prediction : 3D or 4D Niimg
        Same shape as `ground_truth`
    masker: instance of NiftiMasker or MultiNiftiMasker
        Masker to be used on ground_truth and prediction. For more information see:
        http://nilearn.github.io/manipulating_images/masker_objects.html
    loss : str in ['R2', 'corr', 'n_reconstruction_err']
        The loss function used in scoring. Default is normalized
        reconstruction error.
        'R2' :
            The R2 distance between source and target arrays.
            Best possible score is 1.0 and it can be negative (because the
            model can be arbitrarily worse).
        'corr' :
            The correlation between source and target arrays.
        'n_reconstruction_err' :
            The normalized reconstruction error. A perfect prediction
            yields a value of 1.0
    multioutput: str in ['raw_values', 'uniform_average']
        Defines method for aggregating multiple output scores. Default method
        is 'raw_values' i.e. no aggregation.
        'raw_values' :
            Returns a full set of scores in case of multioutput input.
        'uniform_average' :
            Scores of all outputs are averaged with uniform weight.

    Returns
    -------
    score : float or ndarray of floats
        The score or ndarray of scores if ‘multioutput’ is ‘raw_values’.
        The worst possible score is arbitrarily set to -1 for all metrics.
    """
    X_gt = masker.transform(ground_truth)
    X_pred = masker.transform(prediction)

    if loss is "R2":
        score = r2_score(X_gt, X_pred, multioutput=multioutput)
    elif loss is "n_reconstruction_err":
        score = normalized_reconstruction_error(
            X_gt, X_pred, multioutput=multioutput)
    elif loss is "corr":
        score = np.array([pearsonr(X_gt[:, vox], X_pred[:, vox])[0]  # pearsonr returns both rho and p
                          for vox in range(X_pred.shape[1])])
        if multioutput == "uniform_average":
            score = np.mean(score)
    else:
        raise NameError(
            "Unknown loss. Recognized values are 'R2', 'corr', or 'reconstruction_err'")
    # if the calculated score is less than -1, return -1
    return np.maximum(score, -1)


def normalized_reconstruction_error(y_true, y_pred, sample_weights=None,
                                    multioutput='raw_values'):
    """
    Calculates the normalized reconstruction error
    as defined by Bazeille and colleagues (2019).

    A perfect prediction yields a value of 1.

    Parameters
    ----------
    y_true : arr
        The ground truth array.
    y_pred : arr
        The predicted array.
    sample_weights : arr
        Weights to assign to each sample.
    multioutput: str in ['raw_values', 'uniform_average']
        Defines method for aggregating multiple output scores. Default method
        is 'raw_values' i.e. no aggregation.
        'raw_values' :
            Returns a full set of scores in case of multioutput input.
        'uniform_average' :
            Scores of all outputs are averaged with uniform weight.

    Returns
    -------
    score : float or ndarray of floats
        The score or ndarray of scores if `multioutput` is `raw_values`.

    References
    ----------
    `Bazeille T., Richard H., Janati H., and Thirion B. (2019) Local
    Optimal Transport for Functional Brain Template Estimation.
    In: Chung A., Gee J., Yushkevich P., and Bao S. (eds) Information
    Processing in Medical Imaging. Lecture Notes in Computer Science,
    vol 11492. Springer, Cham.
    DOI: 10.1007/978-3-030-20351-1_18.`
    """
    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))

    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))

    numerator = ((y_true - y_pred) ** 2).sum(axis=0, dtype=np.float64)
    denominator = ((y_true) ** 2).sum(axis=0, dtype=np.float64)

    # Include only non-zero values
    nonzero_denominator = (denominator != 0)
    nonzero_numerator = (numerator != 0)
    valid_score = (nonzero_denominator & nonzero_numerator)

    # Calculate reconstruction error
    output_scores = np.ones([y_true.shape[-1]])
    output_scores[valid_score] = 1 - (numerator[valid_score] /
                                      denominator[valid_score])
    if multioutput == 'raw_values':
        # return scores individually
        return output_scores

    elif multioutput == 'uniform_average':
        # passing None as weights yields uniform average
        return np.average(output_scores, weights=None)


def reconstruction_ratio(aligned_error, identity_error):
    """
    Calculates the reconstruction error
    as defined by Bazeille and
    colleagues (2019).

    A value greater than 0 indicates that
    voxels are predicted better by aligned data
    than by raw data.

    Parameters
    ----------
    aligned_error : float or ndarray of floats
        The reconstruction error from a given
        functional alignment method
    identity error :  float or ndarray of floats
        The reconstruction error from predicting
        the target subject as the source subject

    References
    ----------
    `Bazeille T., Richard H., Janati H., and Thirion B. (2019) Local
    Optimal Transport for Functional Brain Template Estimation.
    In: Chung A., Gee J., Yushkevich P., and Bao S. (eds) Information
    Processing in Medical Imaging. Lecture Notes in Computer Science,
    vol 11492. Springer, Cham.
    DOI: 10.1007/978-3-030-20351-1_18.`
    """
    num = 1 - aligned_error
    den = 1 - identity_error
    try:
        return 1 - (num / den)
    except ZeroDivisionError:
        return 0.0



# -*- coding: utf-8 -*-
""" Module for functional template inference using functional alignment on Niimgs and
prediction of new subjects unseen images
"""
# Author: T. Bazeille, B. Thirion
# License: simplified BSD

from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
from joblib import (delayed, Memory, Parallel)
from nilearn.image import index_img, concat_imgs, load_img
from nilearn.input_data.masker_validation import check_embedded_nifti_masker
from fmralign.pairwise_alignment import PairwiseAlignment


def _rescaled_euclidean_mean(imgs, masker, scale_average=False):
    """ Make the Euclidian average of images

    Parameters
    ----------
    imgs: list of Niimgs
        Each img is 3D by default, but can also be 4D.
    masker: instance of NiftiMasker or MultiNiftiMasker
        Masker to be used on the data.
    scale_average: boolean
        If true, the returned average is scaled to have the average norm of imgs
        If false, it will usually have a smaller norm than initial average
        because noise will cancel across images

    Returns
    -------
    average_img: Niimg
        Average of imgs, with same shape as one img
    """
    masked_imgs = [masker.transform(img) for img in imgs]
    average_img = np.mean(masked_imgs, axis=0)
    scale = 1
    if scale_average:
        X_norm = 0
        for img in masked_imgs:
            X_norm += np.linalg.norm(img)
        X_norm /= len(masked_imgs)
        scale = X_norm / np.linalg.norm(average_img)
    average_img *= scale

    return masker.inverse_transform(average_img)


def _align_images_to_template(imgs, template, alignment_method,
                              n_pieces, clustering, n_bags, masker,
                              memory, memory_level, n_jobs, verbose):
    '''Convenience function : for a list of images, return the list
    of estimators (PairwiseAlignment instances) aligning each of them to a
    common target, the template. All arguments are used in PairwiseAlignment
    '''
    aligned_imgs = []
    for img in imgs:
        piecewise_estimator = \
            PairwiseAlignment(n_pieces=n_pieces,
                              alignment_method=alignment_method,
                              clustering=clustering, n_bags=n_bags,
                              mask=masker, memory=memory,
                              memory_level=memory_level,
                              n_jobs=n_jobs,
                              verbose=verbose)
        piecewise_estimator.fit(img, template)
        aligned_imgs.append(piecewise_estimator.transform(img))
    return aligned_imgs


def _create_template(imgs, n_iter, scale_template, alignment_method, n_pieces,
                     clustering, n_bags, masker, memory, memory_level,
                     n_jobs, verbose):
    '''Create template through alternate minimization.  Compute iteratively :
        * T minimizing sum(||R_i X_i-T||) which is the mean of aligned images (RX_i)
        * align initial images to new template T
            (find transform R_i minimizing ||R_i X_i-T|| for each img X_i)


        Parameters
        ----------
        imgs: List of Niimg-like objects
           See http://nilearn.github.io/manipulating_images/input_output.html
           source data. Every img must have the same length (n_sample)
        scale_template: boolean
            If true, template is rescaled after each inference so that it keeps
            the same norm as the average of training images.
        n_iter: int
           Number of iterations in the alternate minimization. Each image is
           aligned n_iter times to the evolving template. If n_iter = 0,
           the template is simply the mean of the input images.
        All other arguments are the same are passed to PairwiseAlignment

        Returns
        -------
        template: list of 3D Niimgs of length (n_sample)
            Models the barycenter of input imgs
        template_history: list of list of 3D Niimgs
            List of the intermediate templates computed at the end of each iteration
    '''

    aligned_imgs = imgs
    template_history = []
    for iter in range(n_iter):
        template = _rescaled_euclidean_mean(
            aligned_imgs, masker, scale_template)
        if 0 < iter < n_iter - 1:
            template_history.append(template)
        aligned_imgs = _align_images_to_template(imgs, template,
                                                 alignment_method, n_pieces,
                                                 clustering, n_bags,
                                                 masker, memory, memory_level,
                                                 n_jobs, verbose)

    return template, template_history


def _map_template_to_image(imgs, train_index, template, alignment_method,
                           n_pieces, clustering, n_bags, masker,
                           memory, memory_level, n_jobs, verbose):
    '''Learn alignment operator from the template toward new images.

    Parameters
    ----------
    imgs: list of 3D Niimgs
        Target images to learn mapping from the template to a new subject
    train_index: list of int
        Matching index between imgs and the corresponding template images to use
        to learn alignment. len(train_index) must be equal to len(imgs)
    template: list of 3D Niimgs
        Learnt in a first step now used as source image
    All other arguments are the same are passed to PairwiseAlignment


    Returns
    -------
    mapping: instance of PairwiseAlignment class
        Alignment estimator fitted to align the template with the input images
    '''

    mapping_image = index_img(template, train_index)
    mapping = PairwiseAlignment(n_pieces=n_pieces,
                                alignment_method=alignment_method,
                                clustering=clustering,
                                n_bags=n_bags, mask=masker, memory=memory,
                                memory_level=memory_level,
                                n_jobs=n_jobs, verbose=verbose)
    mapping.fit(mapping_image, imgs)
    return mapping


def _predict_from_template_and_mapping(template, test_index, mapping):
    """ From a template, and an alignment estimator, predict new contrasts

    Parameters
    ----------
    template: list of 3D Niimgs
        Learnt in a first step now used to predict some new data
    test_index:
        Index of the images not used to learn the alignment mapping and so
        predictable without overfitting
    mapping: instance of PairwiseAlignment class
        Alignment estimator that must have been fitted already

    Returns
    -------
    transformed_image: list of Niimgs
        Prediction corresponding to each template image with index in test_index
        once realigned to the new subjects
    """
    image_to_transform = index_img(template, test_index)
    transformed_image = mapping.transform(image_to_transform)
    return transformed_image


class TemplateAlignment(BaseEstimator, TransformerMixin):
    """
    Decompose the source images into regions and summarize subjects information \
    in a template, then use pairwise alignment to predict \
    new contrast for target subject.
    """

    def __init__(self, alignment_method="identity", n_pieces=1,
                 clustering='kmeans', scale_template=False,
                 n_iter=2, save_template=None, n_bags=1,
                 mask=None, smoothing_fwhm=None, standardize=False,
                 detrend=None, target_affine=None, target_shape=None,
                 low_pass=None, high_pass=None, t_r=None,
                 memory=Memory(cachedir=None), memory_level=0,
                 n_jobs=1, verbose=0):
        '''
        Parameters
        ----------
        alignment_method: string
            Algorithm used to perform alignment between X_i and Y_i :
            * either 'identity', 'scaled_orthogonal', 'optimal_transport',
            'ridge_cv', 'permutation', 'diagonal'
            * or an instance of one of alignment classes (imported from
            functional_alignment.alignment_methods)
        n_pieces: int, optional (default = 1)
            Number of regions in which the data is parcellated for alignment.
            If 1 the alignment is done on full scale data.
            If > 1, the voxels are clustered and alignment is performed on each
            cluster applied to X and Y.
        clustering : string or 3D Niimg optional (default : kmeans)
            'kmeans', 'ward', 'rena', 'hierarchical_kmeans' method used for
            clustering of voxels based on functional signal,
            passed to nilearn.regions.parcellations
            If 3D Niimg, image used as predefined clustering,
            n_bags and n_pieces are then ignored.
        scale_template: boolean, default False
            rescale template after each inference so that it keeps
            the same norm as the average of training images.
        n_iter: int
           number of iteration in the alternate minimization. Each img is
           aligned n_iter times to the evolving template. If n_iter = 0,
           the template is simply the mean of the input images.
        save_template: None or string(optional)
            If not None, path to which the template will be saved.
        n_bags: int, optional (default = 1)
            If 1 : one estimator is fitted.
            If >1 number of bagged parcellations and estimators used.
        mask: Niimg-like object, instance of NiftiMasker or
                                MultiNiftiMasker, optional (default = None)
            Mask to be used on data. If an instance of masker is passed, then its
            mask will be used. If no mask is given, it will be computed
            automatically by a MultiNiftiMasker with default parameters.
        smoothing_fwhm: float, optional (default = None)
            If smoothing_fwhm is not None, it gives the size in millimeters
            of the spatial smoothing to apply to the signal.
        standardize: boolean, optional (default = None)
            If standardize is True, the time-series are centered and normed:
            their variance is put to 1 in the time dimension.
        detrend: boolean, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details
        target_affine: 3x3 or 4x4 matrix, optional (default = None)
            This parameter is passed to nilearn.image.resample_img.
            Please see the related documentation for details.
        target_shape: 3-tuple of integers, optional (default = None)
            This parameter is passed to nilearn.image.resample_img.
            Please see the related documentation for details.
        low_pass: None or float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        high_pass: None or float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        t_r: float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        memory: instance of joblib.Memory or string (default = None)
            Used to cache the masking process and results of algorithms.
            By default, no caching is done. If a string is given, it is the
            path to the caching directory.
        memory_level: integer, optional (default = None)
            Rough estimator of the amount of memory used by caching.
            Higher value means more memory for caching.
        n_jobs: integer, optional (default = 1)
            The number of CPUs to use to do the computation. -1 means
            'all CPUs', -2 'all CPUs but one', and so on.
        verbose: integer, optional (default = 0)
            Indicate the level of verbosity. By default, nothing is printed.
        '''
        self.template = None
        self.template_history = None
        self.alignment_method = alignment_method
        self.n_pieces = n_pieces
        self.clustering = clustering
        self.n_iter = n_iter
        self.scale_template = scale_template
        self.save_template = save_template
        self.n_bags = n_bags
        self.mask = mask
        self.smoothing_fwhm = smoothing_fwhm
        self.standardize = standardize
        self.detrend = detrend
        self.target_affine = target_affine
        self.target_shape = target_shape
        self.low_pass = low_pass
        self.high_pass = high_pass
        self.t_r = t_r
        self.memory = memory
        self.memory_level = memory_level
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, imgs):
        """
        Learn a template from source images, using alignment.

        Parameters
        ----------
        imgs: List of 4D Niimg-like or List of lists of 3D Niimg-like
            Source subjects data. Each element of the parent list is one subject
            data, and all must have the same length (n_samples).

        Returns
        -------
        self

        Attributes
        ----------
        self.template: 4D Niimg object
            Length : n_samples

        """
        # Check if the input is a list, if list of lists, concatenate each subjects
        # data into one unique image.
        if not isinstance(imgs, (list, np.ndarray)) or len(imgs) < 2:
            raise InputError('The method TemplateAlignment.fit() need a list input. \
                             Each element of the list (Niimg-like or list of Niimgs) \
                             is the data for one subject.')
        else:
            if isinstance(imgs[0], (list, np.ndarray)):
                imgs = [concat_imgs(img) for img in imgs]

        self.masker_ = check_embedded_nifti_masker(self)
        self.masker_.n_jobs = self.n_jobs  # self.n_jobs

        # if masker_ has been provided a mask_img
        if self.masker_.mask_img is None:
            self.masker_.fit(imgs)
        else:
            self.masker_.fit()

        self.template, self.template_history = \
            _create_template(imgs, self.n_iter, self.scale_template,
                             self.alignment_method, self.n_pieces,
                             self.clustering, self.n_bags,
                             self.masker_, self.memory, self.memory_level,
                             self.n_jobs, self.verbose)
        if self.save_template is not None:
            self.template.to_filename(self.save_template)

    def transform(self, imgs, train_index, test_index):
        """ Learn alignment between new subject and template calculated during fit,
        then predicts other conditions for this new subject.
        Alignment is learnt between imgs and conditions in the template indexed by train_index.
        Prediction correspond to conditions in the template index by test_index.

        Parameters
        ----------
        imgs: List of 3D Niimg-like objects
            Target subjects known data. Every img must have length (number of sample) train_index.
        train_index: list of ints
            Indexes of the 3D samples used to map each img to the template.
            Every index should be smaller than the number of images in the template.
        test_index: list of ints
            Indexes of the 3D samples to predict from the template and the mapping.
            Every index should be smaller than the number of images in the template.


        Returns
        -------
        predicted_imgs: List of 3D Niimg-like objects
            Target subjects predicted data. Each Niimg has the same length as the list test_index

        """
        if not isinstance(imgs, (list, np.ndarray)):
            raise InputError('The method TemplateAlignment.transform() need a list input. \
                             Each element of the list (Niimg-like or list of Niimgs) \
                             is the data used to align one new subject with images \
                             indexed by train_index.')
        else:
            if isinstance(imgs[0], (list, np.ndarray)) and len(imgs[0]) != len(train_index):
                raise ValueError(' Each element of imgs (Niimg-like or list of Niimgs) \
                                 should have the same length as the length of train_index.')
            elif load_img(imgs[0]).shape[-1] != len(train_index):
                raise ValueError(
                    ' Each element of imgs (Niimg-like or list of Niimgs) \
                    should have the same length as the length of train_index.')

        template_length = self.template.shape[-1]
        if not (all(i < template_length for i in test_index) and all(
                i < template_length for i in train_index)):
            raise ValueError(
                "Template has {} images but you provided a greater index in \
                train_index or test_index.".format(template_length))

        fitted_mappings = Parallel(self.n_jobs, prefer="threads", verbose=self.verbose)(
            delayed(_map_template_to_image)
            (img, train_index, self.template, self.alignment_method,
             self.n_pieces, self.clustering, self.n_bags, self.masker_,
             self.memory, self.memory_level, self.n_jobs, self.verbose
             ) for img in imgs
        )

        predicted_imgs = Parallel(self.n_jobs, prefer="threads", verbose=self.verbose)(
            delayed(_predict_from_template_and_mapping)
            (self.template, test_index, mapping) for mapping in fitted_mappings
        )
        return predicted_imgs

    # Make inherited function harmless
    def fit_transform(self):
        """Parent method not applicable here. Will raise AttributeError if called.
        """
        raise AttributeError(
            "type object 'PairwiseAlignment' has no attribute 'fit_transform'")



# -*- coding: utf-8 -*-
""" Module for pairwise functional alignment
"""
import warnings
import numpy as np
import os

from sklearn.base import BaseEstimator, TransformerMixin
from joblib import (delayed, Memory, Parallel)
from sklearn.model_selection import ShuffleSplit
from sklearn.base import clone
from nilearn.input_data.masker_validation import check_embedded_nifti_masker
from nilearn.image import load_img, concat_imgs, index_img
import nibabel as nib
from fmralign.alignment_methods import RidgeAlignment, Identity, Hungarian, \
    ScaledOrthogonalAlignment, OptimalTransportAlignment, DiagonalAlignment
from fmralign._utils import _make_parcellation, piecewise_transform, _intersect_clustering_mask


def generate_Xi_Yi(labels, X, Y, masker, verbose):
    """ Generate source and target data X_i and Y_i for each piece i.

    Parameters
    ----------
    labels : list of ints (len n_features)
        Parcellation of features in clusters
    X: Niimg-like object
        Source data
    Y: Niimg-like object
        Target data
    masker: instance of NiftiMasker or MultiNiftiMasker
        Masker to be used on the data. For more information see:
        http://nilearn.github.io/manipulating_images/masker_objects.html
    verbose: integer, optional.
        Indicate the level of verbosity.

    Yields
    -------
    X_i: ndarray
        Source data for piece i (shape : n_samples, n_features)
    Y_i: ndarray
        Target data for piece i (shape : n_samples, n_features)

    """
    X_ = masker.transform(X)
    Y_ = masker.transform(Y)
    unique_labels = np.unique(labels)

    for k in range(len(unique_labels)):
        label = unique_labels[k]
        i = label == labels
        if (k + 1) % 25 == 0 and verbose > 0:
            print("Fitting parcel: " + str(k + 1) +
                  "/" + str(len(unique_labels)))
        # should return X_i Y_i
        yield X_[:, i], Y_[:, i]


def fit_one_piece(X_i, Y_i, alignment_method):
    """ Align source and target data in one piece i, X_i and Y_i, using
    alignment method and learn transformation to map X to Y.

    Parameters
    ----------
    X_i: ndarray
        Source data for piece i (shape : n_samples, n_features)
    Y_i: ndarray
        Target data for piece i (shape : n_samples, n_features)
    alignment_method: string
        Algorithm used to perform alignment between X_i and Y_i :
        - either 'identity', 'scaled_orthogonal', 'optimal_transport',
        'ridge_cv', 'permutation', 'diagonal'
        - or an instance of one of alignment classes
            (imported from functional_alignment.alignment_methods)
    Returns
    -------
    alignment_algo
        Instance of alignment estimator class fitted for X_i, Y_i
    """

    if alignment_method == 'identity':
        alignment_algo = Identity()
    elif alignment_method == 'scaled_orthogonal':
        alignment_algo = ScaledOrthogonalAlignment()
    elif alignment_method == 'ridge_cv':
        alignment_algo = RidgeAlignment()
    elif alignment_method == 'permutation':
        alignment_algo = Hungarian()
    elif alignment_method == 'optimal_transport':
        alignment_algo = OptimalTransportAlignment()
    elif alignment_method == 'diagonal':
        alignment_algo = DiagonalAlignment()
    elif isinstance(alignment_method, (Identity, ScaledOrthogonalAlignment,
                                       RidgeAlignment, Hungarian,
                                       OptimalTransportAlignment,
                                       DiagonalAlignment)):
        alignment_algo = clone(alignment_method)

    if not np.count_nonzero(X_i) or not np.count_nonzero(Y_i):
        warn_msg = ("Empty parcel found. Please check overlap between " +
                    "provided mask and functional image. Returning " +
                    "Identity alignment for empty parcel")
        warnings.warn(warn_msg)
        alignment_algo = Identity()
    try:
        alignment_algo.fit(X_i, Y_i)
    except UnboundLocalError:
        warn_msg = ("{} is an unrecognized ".format(alignment_method) +
                    "alignment method. Please provide a recognized " +
                    "alignment method.")
        raise NotImplementedError(warn_msg)
    return alignment_algo


def fit_one_parcellation(X_, Y_, alignment_method, masker, n_pieces,
                         clustering, clustering_index,
                         n_jobs, verbose):
    """ Create one parcellation of n_pieces and align each source and target
    data in one piece i, X_i and Y_i, using alignment method
    and learn transformation to map X to Y.

    Parameters
    ----------
    X_: Niimg-like object
        Source data
    Y_: Niimg-like object
        Target data
    alignment_method: string
        algorithm used to perform alignment between each region of X_ and Y_
    masker: instance of NiftiMasker or MultiNiftiMasker
        Masker to be used on the data. For more information see:
        http://nilearn.github.io/manipulating_images/masker_objects.html
    n_pieces: n_pieces: int,
        Number of regions in which the data is parcellated for alignment
    clustering: string or 3D Niimg
        method used to perform parcellation of data.
        If 3D Niimg, image used as predefined clustering.
    clustering_index: list of integers
        Clustering is performed on a 20% subset of the data chosen randomly
        in timeframes. This index carry this subset.
    n_jobs: integer, optional
        The number of CPUs to use to do the computation. -1 means
        'all CPUs', -2 'all CPUs but one', and so on.
    verbose: integer, optional
        Indicate the level of verbosity. By default, nothing is printed

    Returns
    -------
    alignment_algo
        Instance of alignment estimator class fitted for X_i, Y_i
    """
    # choose indexes maybe with index_img to not
    labels = _make_parcellation(X_, clustering_index, clustering,
                                n_pieces, masker, verbose=verbose)

    fit = Parallel(n_jobs, prefer="threads", verbose=verbose)(
        delayed(fit_one_piece)(
            X_i, Y_i, alignment_method
        ) for X_i, Y_i in generate_Xi_Yi(labels, X_, Y_, masker, verbose)
    )

    return labels, fit


class PairwiseAlignment(BaseEstimator, TransformerMixin):
    """
    Decompose the source and target images into regions and align corresponding \
    regions independently.
    """

    def __init__(self, alignment_method, n_pieces=1,
                 clustering='kmeans', n_bags=1, mask=None,
                 smoothing_fwhm=None, standardize=False, detrend=False,
                 target_affine=None, target_shape=None, low_pass=None,
                 high_pass=None, t_r=None,
                 memory=Memory(cachedir=None), memory_level=0,
                 n_jobs=1, verbose=0):
        """
        If n_pieces > 1, decomposes the images into regions \
        and align each source/target region independantly.
        If n_bags > 1, this parcellation process is applied multiple time \
        and the resulting models are bagged.

        Parameters
        ----------
        alignment_method: string
            Algorithm used to perform alignment between X_i and Y_i :
            * either 'identity', 'scaled_orthogonal', 'optimal_transport',
            'ridge_cv', 'permutation', 'diagonal'
            * or an instance of one of alignment classes
            (imported from functional_alignment.alignment_methods)
        n_pieces: int, optional (default = 1)
            Number of regions in which the data is parcellated for alignment.
            If 1 the alignment is done on full scale data.
            If >1, the voxels are clustered and alignment is performed
            on each cluster applied to X and Y.
        clustering : string or 3D Niimg optional (default : kmeans)
            'kmeans', 'ward', 'rena', 'hierarchical_kmeans' method used for
            clustering of voxels based on functional signal, passed to
            nilearn.regions.parcellations
            If 3D Niimg, image used as predefined clustering,
            n_bags and n_pieces are then ignored.
        n_bags: int, optional (default = 1)
            If 1 : one estimator is fitted.
            If >1 number of bagged parcellations and estimators used.
        mask: Niimg-like object, instance of NiftiMasker or
                                MultiNiftiMasker, optional (default = None)
            Mask to be used on data. If an instance of masker is passed,
            then its mask will be used. If no mask is given,
            it will be computed automatically by a MultiNiftiMasker
            with default parameters.
        smoothing_fwhm: float, optional (default = None)
            If smoothing_fwhm is not None, it gives the size in millimeters
            of the spatial smoothing to apply to the signal.
        standardize: boolean, optional (default = False)
            If standardize is True, the time-series are centered and normed:
            their variance is put to 1 in the time dimension.
        detrend: boolean, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details
        target_affine: 3x3 or 4x4 matrix, optional (default = None)
            This parameter is passed to nilearn.image.resample_img.
            Please see the related documentation for details.
        target_shape: 3-tuple of integers, optional (default = None)
            This parameter is passed to nilearn.image.resample_img.
            Please see the related documentation for details.
        low_pass: None or float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        high_pass: None or float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        t_r: float, optional (default = None)
            This parameter is passed to nilearn.signal.clean.
            Please see the related documentation for details.
        memory: instance of joblib.Memory or string (default = None)
            Used to cache the masking process and results of algorithms.
            By default, no caching is done. If a string is given, it is the
            path to the caching directory.
        memory_level: integer, optional (default = None)
            Rough estimator of the amount of memory used by caching.
            Higher value means more memory for caching.
        n_jobs: integer, optional (default = 1)
            The number of CPUs to use to do the computation. -1 means
            'all CPUs', -2 'all CPUs but one', and so on.
        verbose: integer, optional (default = 0)
            Indicate the level of verbosity. By default, nothing is printed.
        """
        self.n_pieces = n_pieces
        self.alignment_method = alignment_method
        self.n_bags = n_bags
        self.clustering = clustering
        self.mask = mask
        self.smoothing_fwhm = smoothing_fwhm
        self.standardize = standardize
        self.detrend = detrend
        self.low_pass = low_pass
        self.high_pass = high_pass
        self.t_r = t_r
        self.target_affine = target_affine
        self.target_shape = target_shape
        self.memory = memory
        self.memory_level = memory_level
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X, Y):
        """Fit data X and Y and learn transformation to map X to Y

        Parameters
        ----------
        X: Niimg-like object
            Source data.

        Y: Niimg-like object
            Target data

        Returns
        -------
        self
        """
        self.masker_ = check_embedded_nifti_masker(self)
        self.masker_.n_jobs = self.n_jobs

        if self.masker_.mask_img is None:
            self.masker_.fit([X])
        else:
            self.masker_.fit()

        if type(self.clustering) == nib.nifti1.Nifti1Image or os.path.isfile(self.clustering):
            # check that clustering provided fills the mask, if not, reduce the mask
            if 0 in self.masker_.transform(self.clustering):
                reduced_mask = _intersect_clustering_mask(
                    self.clustering, self.masker_.mask_img)
                self.mask = reduced_mask
                self.masker_ = check_embedded_nifti_masker(self)
                self.masker_.n_jobs = self.n_jobs
                self.masker_.fit()
                warnings.warn(
                    "Mask used was bigger than clustering provided. " +
                    "Its intersection with the clustering was used instead.")

        if isinstance(X, (list, np.ndarray)):
            X_ = concat_imgs(X)
        else:
            X_ = load_img(X)
        if isinstance(X, (list, np.ndarray)):
            Y_ = concat_imgs(Y)
        else:
            Y_ = load_img(Y)

        self.fit_, self.labels_ = [], []
        rs = ShuffleSplit(n_splits=self.n_bags,
                          test_size=.8, random_state=0)

        outputs = Parallel(n_jobs=self.n_jobs, prefer="threads",
                           verbose=self.verbose)(
            delayed(fit_one_parcellation)(
                X_, Y_, self.alignment_method, self.masker_, self.n_pieces,
                self.clustering, clustering_index, self.n_jobs, self.verbose)
            for clustering_index, _ in rs.split(range(X_.shape[-1])))
        # change split
        self.labels_ = [output[0] for output in outputs]
        self.fit_ = [output[1] for output in outputs]

        return self

    def transform(self, X):
        """Predict data from X

        Parameters
        ----------
        X: Niimg-like object
            Source data

        Returns
        -------
        X_transform: Niimg-like object
            Predicted data
        """
        if isinstance(X, (list, np.ndarray)):
            X = concat_imgs(X)
        X_ = self.masker_.transform(X)

        X_transform = np.zeros_like(X_)
        for i in range(self.n_bags):
            X_transform += piecewise_transform(
                self.labels_[i], self.fit_[i], X_)

        X_transform /= self.n_bags

        return self.masker_.inverse_transform(X_transform)

    # Make inherited function harmless
    def fit_transform(self):
        """Parent method not applicable here. Will raise AttributeError if called.
        """
        raise AttributeError(
            "type object 'PairwiseAlignment' has no attribute 'fit_transform'")



# -*- coding: utf-8 -*-
import copy
import pytest
import numpy as np
from sklearn.utils.testing import assert_greater
from nilearn.input_data import NiftiMasker
from nilearn.image import new_img_like
from fmralign.pairwise_alignment import PairwiseAlignment, fit_one_piece
from fmralign.tests.utils import (assert_algo_transform_almost_exactly,
                                  zero_mean_coefficient_determination,
                                  random_niimg)


def test_unsupported_alignment():
    img1, mask_img = random_niimg((8, 7, 6, 10))
    img2, _ = random_niimg((7, 6, 8, 5))
    args = {'alignment_method': 'scaled_procrustes', 'mask': mask_img}
    algo = PairwiseAlignment(**args)
    with pytest.raises(NotImplementedError):
        algo.fit(img1, img2)


def test_optimal_transport_error_big_parcels():
    n_voxels, n_features = 6000, 10
    X, Y = np.ones((n_features, n_voxels)), np.ones((n_features, n_voxels))
    with pytest.warns(UserWarning):
        estimator = fit_one_piece(X, Y, "optimal_transport")


def test_pairwise_identity():
    img1, mask_img = random_niimg((8, 7, 6, 10))

    args_list = [{'alignment_method': 'identity', 'mask': mask_img},
                 {'alignment_method': 'identity', 'n_pieces': 3,
                  'mask': mask_img},
                 {'alignment_method': 'identity', 'n_pieces': 3,
                     'n_bags': 4, 'mask': mask_img},
                 {'alignment_method': 'identity', 'n_pieces': 3,
                     'n_bags': 3, 'mask': mask_img, 'n_jobs': 2}
                 ]
    for args in args_list:
        algo = PairwiseAlignment(**args)
        assert_algo_transform_almost_exactly(
            algo, img1, img1, mask=mask_img)

    # test intersection of clustering and mask
    data_mask = copy.deepcopy(mask_img.get_fdata())
    data_mask[0] = 0
    # create ground truth
    clustering_mask = new_img_like(mask_img, data_mask)
    data_clust = copy.deepcopy(data_mask)
    data_clust[1] = 3
    # create 2-parcels clustering, smaller than background
    clustering = new_img_like(mask_img, data_clust)

    # clustering is smaller than mask
    assert (mask_img.get_fdata() > 0).sum() > (clustering.get_data() > 0).sum()
    algo = PairwiseAlignment(alignment_method='identity',
                             mask=mask_img, clustering=clustering)
    with pytest.warns(UserWarning):
        algo.fit(img1, img1)
    assert (algo.mask.get_fdata() > 0).sum() == (
        clustering.get_fdata() > 0).sum()

    # test warning raised if parcel is 0 :
    null_im = new_img_like(img1, np.zeros_like(img1.get_fdata()))
    with pytest.warns(UserWarning):
        algo.fit(null_im, null_im)


def test_models_against_identity():
    img1, mask_img = random_niimg((7, 6, 8, 5))
    img2, _ = random_niimg((7, 6, 8, 5))
    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()
    ground_truth = masker.transform(img2)
    identity_baseline_score = zero_mean_coefficient_determination(
        ground_truth, masker.transform(img1))
    for alignment_method in ['permutation',  'ridge_cv', 'scaled_orthogonal',
                             'optimal_transport', 'diagonal']:
        for clustering in ["kmeans", "hierarchical_kmeans"]:
            algo = PairwiseAlignment(
                alignment_method=alignment_method, mask=masker,
                clustering=clustering, n_pieces=2, n_bags=1, n_jobs=1)
            algo.fit(img1, img2)
            im_test = algo.transform(img1)
            algo_score = zero_mean_coefficient_determination(ground_truth,
                                                             masker.transform(
                                                                 im_test))
            assert_greater(algo_score, identity_baseline_score)



# -*- coding: utf-8 -*-
import os
import numpy as np
import nibabel as nib
from packaging import version
from scipy.stats import pearsonr
from sklearn.cluster import MiniBatchKMeans
import nilearn
from nilearn.image import smooth_img, index_img, new_img_like
from nilearn.regions.parcellations import Parcellations
from nilearn.masking import _apply_mask_fmri, intersect_masks
from nilearn._utils.niimg_conversions import _check_same_fov
import warnings


def _intersect_clustering_mask(clustering, mask):
    "Take 3D Niimg clustering and bigger mask, output reduced mask"
    dat = clustering.get_data()
    new_ = np.zeros_like(dat)
    new_[dat > 0] = 1
    clustering_mask = new_img_like(clustering, new_)
    return intersect_masks(
        [clustering_mask, mask], threshold=1, connected=True)


def piecewise_transform(labels, estimators, X):
    """ Apply a piecewise transform to X:
    Parameters
    ----------
    labels: list of ints (len n_features)
        Parcellation of features in clusters
    estimators: list of estimators with transform() method
        I-th estimator will be applied on the i-th cluster of features
    X: nd array (n_samples, n_features)
        Data to transform

    Returns
    -------
    X_transform: nd array (n_features, n_samples)
        Transformed data
    """
    unique_labels = np.unique(labels)
    X_transform = np.zeros_like(X)

    for i in range(len(unique_labels)):
        label = unique_labels[i]
        X_transform[:, labels == label] = estimators[i].transform(
            X[:, labels == label])
    return X_transform


def _remove_empty_labels(labels):
    '''Remove empty values label values from labels list'''
    vals = np.unique(labels)
    inverse_vals = - np.ones(labels.max() + 1).astype(np.int)
    inverse_vals[vals] = np.arange(len(vals))
    return inverse_vals[labels]


def _check_labels(labels, threshold=1000):
    """ Check is some parcels are bigger than a certain threshold and raise warning if so
    """
    unique_labels, counts = np.unique(
        labels, return_counts=True)

    if not all(count < threshold for count in counts):
        warning = "\n Some parcels are more than 1000 voxels wide it can slow down alignment, especially optimal_transport :"
        for i in range(len(counts)):
            if counts[i] > threshold:
                warning += "\n parcel {} : {} voxels".format(
                    unique_labels[i], counts[i])
        warnings.warn(warning)
    pass


def _hierarchical_k_means(X, n_clusters, init="k-means++", batch_size=1000,
                          n_init=10, max_no_improvement=10, verbose=0, random_state=0):
    """ Use a recursive k-means to cluster X
    Parameters
    ----------
    X: nd array (n_samples, n_features)
        Data to cluster

    n_clusters: int,
        The number of clusters to find.

    init : {'k-means++', 'random' or an ndarray}
        Method for initialization, defaults to 'k-means++':
        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.
        'random': choose k observations (rows) at random from data for
        the initial centroids.
        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centers.

    batch_size : int, optional, default: 100
        Size of the mini batches. (Kmeans performed through MiniBatchKMeans)

    n_init : int, default=3
        Number of random initializations that are tried.
        In contrast to KMeans, the algorithm is only run once, using the
        best of the ``n_init`` initializations as measured by inertia.

    max_no_improvement : int, default: 10
        Control early stopping based on the consecutive number of mini
        batches that does not yield an improvement on the smoothed inertia.
        To disable convergence detection based on inertia, set
        max_no_improvement to None.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for centroid initialization and
        random reassignment. Use an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.
    Returns
    -------
    labels : list of ints (len n_features)
        Parcellation of features in clusters
    """

    n_big_clusters = int(np.sqrt(n_clusters))
    mbk = MiniBatchKMeans(init=init, n_clusters=n_big_clusters, batch_size=batch_size,
                          n_init=n_init, max_no_improvement=max_no_improvement, verbose=verbose,
                          random_state=random_state).fit(X)
    coarse_labels = mbk.labels_
    fine_labels = np.zeros_like(coarse_labels)
    q = 0
    for i in range(n_big_clusters):
        n_small_clusters = int(
            n_clusters * np.sum(coarse_labels == i) * 1. / X.shape[0])
        n_small_clusters = np.maximum(1, n_small_clusters)
        mbk = MiniBatchKMeans(init=init, n_clusters=n_small_clusters,
                              batch_size=batch_size, n_init=n_init,
                              max_no_improvement=max_no_improvement, verbose=verbose,
                              random_state=random_state).fit(X[coarse_labels == i])
        fine_labels[coarse_labels == i] = q + mbk.labels_
        q += n_small_clusters

    return _remove_empty_labels(fine_labels)


def _make_parcellation(imgs, clustering_index, clustering, n_pieces, masker, smoothing_fwhm=5, verbose=0):
    """Convenience function to use nilearn Parcellation class in our pipeline.
    It is used to find local regions of the brain in which alignment will be later applied.
    For alignment computational efficiency, regions should be of hundreds of voxels.

    Parameters
    ----------
    imgs: Niimgs
        data to cluster
    clustering_index: list of integers
        Clustering is performed on a subset of the data chosen randomly
        in timeframes. This index carries this subset.
    clustering: string or 3D Niimg
        In : {'kmeans', 'ward', 'rena'}, passed to nilearn Parcellations class.
        If you aim for speed, choose k-means (and check kmeans_smoothing_fwhm parameter)
        If you want spatially connected and/or reproducible regions use 'ward'
        If you want balanced clusters (especially from timeseries) used 'hierarchical_kmeans'
        For 'rena', need nilearn > 0.5.2
        If 3D Niimg, image used as predefined clustering, n_pieces is ignored
    n_pieces: int
        number of different labels
    masker: instance of NiftiMasker or MultiNiftiMasker
        Masker to be used on the data. For more information see:
        http://nilearn.github.io/manipulating_images/masker_objects.html
    smoothing_fwhm: None or int
        By default 5mm smoothing will be applied before kmeans clustering to have
        more compact clusters (but this will not change the data later).
        To disable this option, this parameter should be None.

    Returns
    -------
    labels : list of ints (len n_features)
        Parcellation of features in clusters
    """
    # check if clustering is provided
    if type(clustering) == nib.nifti1.Nifti1Image or os.path.isfile(clustering):
        _check_same_fov(masker.mask_img_, clustering)
        labels = _apply_mask_fmri(clustering, masker.mask_img_).astype(int)

    # otherwise check it's needed, if not return 1 everywhere
    elif n_pieces == 1:
        labels = np.ones(
            int(masker.mask_img_.get_fdata().sum()), dtype=np.int8)
    # otherwise check requested clustering method
    elif clustering == "hierarchical_kmeans" and n_pieces > 1:
        imgs_subset = index_img(imgs, clustering_index)
        if smoothing_fwhm is not None:
            X = masker.transform(smooth_img(imgs_subset, smoothing_fwhm))
        else:
            X = masker.transform(imgs_subset)
        labels = _hierarchical_k_means(
            X.T, n_clusters=n_pieces, verbose=verbose) + 1

    elif clustering in ['kmeans', 'ward', 'rena'] and n_pieces > 1:
        imgs_subset = index_img(imgs, clustering_index)
        if clustering == "kmeans" and smoothing_fwhm is not None:
            images_to_parcel = smooth_img(imgs_subset, smoothing_fwhm)
        else:
            images_to_parcel = imgs_subset
        try:
            parcellation = Parcellations(method=clustering, n_parcels=n_pieces, mask=masker,
                                         scaling=False, n_iter=20, verbose=verbose)
        except TypeError:
            if clustering == "rena" and (version.parse(nilearn.__version__) <= version.parse("0.5.2")):
                raise InputError(
                    ('ReNA algorithm is only available in Nilearn version > 0.5.2. \
                     Your version is {}. If you want to use ReNA, please run "pip install nilearn --upgrade"'.format(nilearn.__version__)))
            else:
                parcellation = Parcellations(
                    method=clustering, n_parcels=n_pieces, mask=masker, verbose=verbose)
        parcellation.fit(images_to_parcel)
        labels = _apply_mask_fmri(
            parcellation.labels_img_, masker.mask_img_).astype(int)

    else:
        raise InputError(
            ('Clustering should be "kmeans", "ward", "rena", "hierarchical_kmeans", \
             or a 3D Niimg, and n_pieces should be an integer ≥ 1'))

    if verbose > 0:
        unique_labels, counts = np.unique(
            labels, return_counts=True)
        print("The alignment will be applied on parcels of sizes {}".format(counts))

    # raise warning if some parcels are bigger than 1000 voxels
    _check_labels(labels)

    return labels



# -*- coding: utf-8 -*-
import pytest
from sklearn.utils.testing import assert_array_almost_equal, assert_greater

from nilearn.input_data import NiftiMasker
from nilearn.image import math_img, concat_imgs, index_img

from fmralign.template_alignment import TemplateAlignment, _rescaled_euclidean_mean
from fmralign.tests.utils import random_niimg, zero_mean_coefficient_determination


def test_template_identity():

    n = 10
    im, mask_img = random_niimg((6, 5, 3))

    sub_1 = concat_imgs(n * [im])
    sub_2 = math_img("2 * img", img=sub_1)
    sub_3 = math_img("3 * img", img=sub_1)

    ref_template = sub_2
    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()

    subs = [sub_1, sub_2, sub_3]

    # test euclidian mean function
    euclidian_template = _rescaled_euclidean_mean(subs, masker)
    assert_array_almost_equal(
        ref_template.get_fdata(), euclidian_template.get_fdata())

    # test different fit() accept list of list of 3D Niimgs as input.
    algo = TemplateAlignment(alignment_method='identity', mask=masker)
    algo.fit([n * [im]] * 3)
    # test template
    assert_array_almost_equal(
        sub_1.get_fdata(), algo.template.get_fdata())

    # test fit() transform() with 4D Niimgs input for several params set
    args_list = [{'alignment_method': 'identity', 'mask': masker},
                 {'alignment_method': 'identity', 'mask': masker, 'n_jobs': 2},
                 {'alignment_method': 'identity', 'n_pieces': 3, 'mask': masker},
                 {'alignment_method': 'identity', 'n_pieces': 3,
                     'n_bags': 2, 'mask': masker}
                 ]

    for args in args_list:
        algo = TemplateAlignment(**args)
        # Learning a template which is
        algo.fit(subs)
        # test template
        assert_array_almost_equal(
            ref_template.get_fdata(), algo.template.get_fdata())
        predicted_imgs = algo.transform(
            [index_img(sub_1, range(8))], train_index=range(8),
            test_index=range(8, 10))
        ground_truth = index_img(ref_template, range(8, 10))
        assert_array_almost_equal(
            ground_truth.get_fdata(), predicted_imgs[0].get_fdata())

    # test transform() with wrong indexes length or content (on previous fitted algo)
    train_inds, test_inds = [[0, 1], [1, 10],
                             [4, 11], [0, 1, 2]], [[6, 8, 29], [4, 6], [4, 11], [4, 5]]

    for train_ind, test_ind in zip(train_inds, test_inds):
        with pytest.raises(Exception):
            assert algo.transform(
                [index_img(sub_1, range(2))], train_index=train_ind, test_index=test_ind)

    # test wrong images input in fit() and transform method
    with pytest.raises(Exception):
        assert algo.transform(
            [n * [im]] * 2, train_index=train_inds[-1], test_index=test_inds[-1])
        assert algo.fit([im])
        assert algo.transform(
            [im], train_index=train_inds[-1], test_index=test_inds[-1])


def test_template_closer_to_target():
    n_samples = 6

    subject_1, mask_img = random_niimg((6, 5, 3, n_samples))
    subject_2, _ = random_niimg((6, 5, 3, n_samples))

    masker = NiftiMasker(mask_img=mask_img)
    masker.fit()

    # Calculate metric between each subject and the average

    sub_1 = masker.transform(subject_1)
    sub_2 = masker.transform(subject_2)
    subs = [subject_1, subject_2]
    average_img = _rescaled_euclidean_mean(subs, masker)
    avg_data = masker.transform(average_img)
    mean_distance_1 = zero_mean_coefficient_determination(sub_1, avg_data)
    mean_distance_2 = zero_mean_coefficient_determination(sub_2, avg_data)

    for alignment_method in ['permutation',  'ridge_cv', 'scaled_orthogonal',
                             'optimal_transport', 'diagonal']:
        algo = TemplateAlignment(
            alignment_method=alignment_method, n_pieces=3, n_bags=2, mask=masker)
        # Learn template
        algo.fit(subs)
        # Assess template is closer to mean than both images
        template_data = masker.transform(algo.template)
        template_mean_distance = zero_mean_coefficient_determination(
            avg_data, template_data)
        assert_greater(template_mean_distance, mean_distance_1)
        assert_greater(template_mean_distance, mean_distance_2)



# -*- coding: utf-8 -*-
""" Module implementing alignment estimators on ndarrays
"""

import numpy as np
import scipy
from scipy.spatial.distance import cdist
from scipy import linalg
from scipy.sparse import diags
import sklearn
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.optimize import linear_sum_assignment
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.linear_model import RidgeCV
from joblib import Parallel, delayed
import warnings


def scaled_procrustes(X, Y, scaling=False, primal=None):
    """Compute a mixing matrix R and a scaling sc such that Frobenius norm
    ||sc RX - Y||^2 is minimized and R is an orthogonal matrix.

    Parameters
    ----------
    X: (n_samples, n_features) nd array
        source data
    Y: (n_samples, n_features) nd array
        target data
    scaling: bool
        If scaling is true, computes a floating scaling parameter sc such that:
        ||sc * RX - Y||^2 is minimized and
        - R is an orthogonal matrix
        - sc is a scalar
        If scaling is false sc is set to 1
    primal: bool or None, optional,
         Whether the SVD is done on the YX^T (primal) or Y^TX (dual)
         if None primal is used iff n_features <= n_timeframes

    Returns
    ----------
    R: (n_features, n_features) nd array
        transformation matrix
    sc: int
        scaling parameter
    """
    X = X.astype(np.float64, copy=False)
    Y = Y.astype(np.float64, copy=False)
    if np.linalg.norm(X) == 0 or np.linalg.norm(Y) == 0:
        return np.eye(X.shape[1]), 1
    if primal is None:
        primal = X.shape[0] >= X.shape[1]
    if primal:
        A = Y.T.dot(X)
        if A.shape[0] == A.shape[1]:
            A += + 1.e-18 * np.eye(A.shape[0])
        U, s, V = linalg.svd(A, full_matrices=0)
        R = U.dot(V)
    else:  # "dual" mode
        Uy, sy, Vy = linalg.svd(Y, full_matrices=0)
        Ux, sx, Vx = linalg.svd(X, full_matrices=0)
        A = np.diag(sy).dot(Uy.T).dot(Ux).dot(np.diag(sx))
        U, s, V = linalg.svd(A)
        R = Vy.T.dot(U).dot(V).dot(Vx)

    if scaling:
        sc = s.sum() / (np.linalg.norm(X) ** 2)
    else:
        sc = 1
    return R.T, sc


def optimal_permutation(X, Y):
    """Compute the optmal permutation matrix of X toward Y

    Parameters
    ----------
    X: (n_samples, n_features) nd array
        source data
    Y: (n_samples, n_features) nd array
        target data

    Returns
    ----------
    permutation : (n_features, n_features) nd array
        transformation matrix
    """
    dist = pairwise_distances(X.T, Y.T)
    u = linear_sum_assignment(dist)
    u = np.array(list(zip(*u)))
    permutation = scipy.sparse.csr_matrix(
        (np.ones(X.shape[1]), (u[:, 0], u[:, 1]))).T
    return permutation


def _projection(x, y):
    """Compute scalar d minimizing ||dx-y||

    Parameters
    ----------
    x: (n_features) nd array
        source vector
    y: (n_features) nd array
        target vector

    Returns
    --------
    d: int
        scaling factor
    """
    if (x == 0).all():
        return 0
    else:
        return np.dot(x, y) / np.linalg.norm(x)**2


def _voxelwise_signal_projection(X, Y, n_jobs=1, parallel_backend='threading'):
    """Compute D, list of scalar d_i minimizing :
        ||d_i * x_i - y_i|| for every x_i, y_i in X, Y

    Parameters
    ----------
    X: (n_samples, n_features) nd array
        source data
    Y: (n_samples, n_features) nd array
        target data

    Returns
    --------
    D: list of ints
        List of optimal scaling factors
    """
    return Parallel(n_jobs, parallel_backend)(delayed(_projection)(
        voxel_source, voxel_target)
        for voxel_source, voxel_target in zip(X, Y))


class Alignment(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, Y):
        pass

    def transform(self, X):
        pass


class Identity(Alignment):
    """Compute no alignment, used as baseline for benchmarks : RX = X.
    """

    def transform(self, X):
        """returns X"""
        return X


class DiagonalAlignment(Alignment):
    '''Compute the voxelwise projection factor between X and Y.

    Parameters
    ----------
    n_jobs: integer, optional (default = 1)
        The number of CPUs to use to do the computation. -1 means
        'all CPUs', -2 'all CPUs but one', and so on.
    parallel_backend: str, ParallelBackendBase instance, None (default: 'threading')
        Specify the parallelization backend implementation. For more
        informations see joblib.Parallel documentation

    Attributes
    -----------
    R : scipy.sparse.diags
        Scaling matrix containing the optimal shrinking factor for every voxel
    '''

    def __init__(self, n_jobs=1, parallel_backend='threading'):
        self.n_jobs = n_jobs
        self.parallel_backend = parallel_backend

    def fit(self, X, Y):
        '''

        Parameters
        --------------
        X: (n_samples, n_features) nd array
            source data
        Y: (n_samples, n_features) nd array
            target data'''
        shrinkage_coefficients = _voxelwise_signal_projection(
            X.T, Y.T, self.n_jobs, self.parallel_backend)

        self.R = diags(shrinkage_coefficients)
        return

    def transform(self, X):
        """Transform X using optimal coupling computed during fit.
        """
        return self.R.dot(X.T).T


class ScaledOrthogonalAlignment(Alignment):
    """Compute a orthogonal mixing matrix R and a scaling sc such that Frobenius norm \
    ||sc RX - Y||^2 is minimized.

    Parameters
    -----------
    scaling : boolean, optional
        Determines whether a scaling parameter is applied to improve transform.

    Attributes
    -----------
    R : ndarray (n_features, n_features)
        Optimal orthogonal transform
    """

    def __init__(self, scaling=True):
        self.scaling = scaling
        self.scale = 1

    def fit(self, X, Y):
        """ Fit orthogonal R s.t. ||sc XR - Y||^2

        Parameters
        -----------
        X: (n_samples, n_features) nd array
            source data
        Y: (n_samples, n_features) nd array
            target data
        """
        R, sc = scaled_procrustes(X, Y, scaling=self.scaling)
        self.scale = sc
        self.R = sc * R
        return self

    def transform(self, X):
        """Transform X using optimal transform computed during fit.
        """
        return X.dot(self.R)


class RidgeAlignment(Alignment):
    """ Compute a scikit-estimator R using a mixing matrix M s.t Frobenius \
    norm || XM - Y ||^2 + alpha * ||M||^2 is minimized with cross-validation

    Parameters
    ----------
    R : scikit-estimator from sklearn.linear_model.RidgeCV
        with methods fit, predict
    alpha : numpy array of shape [n_alphas]
        Array of alpha values to try. Regularization strength; \
        must be a positive float. Regularization improves the conditioning \
        of the problem and reduces the variance of the estimates. \
        Larger values specify stronger regularization. Alpha corresponds to \
        ``C^-1`` in other models such as LogisticRegression or LinearSVC.
    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.\
        Possible inputs for cv are:
        -None, to use the efficient Leave-One-Out cross-validation
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.
    """

    def __init__(self, alphas=[0.1, 1.0, 10.0, 100, 1000], cv=4):
        self.alphas = [alpha for alpha in alphas]
        self.cv = cv

    def fit(self, X, Y):
        """ Fit R s.t. || XR - Y ||^2 + alpha ||R||^2 is minimized with cv

        Parameters
        -----------
        X: (n_samples, n_features) nd array
            source data
        Y: (n_samples, n_features) nd array
            target data
        """
        self.R = RidgeCV(alphas=self.alphas, fit_intercept=True,
                         normalize=False,
                         scoring=sklearn.metrics.SCORERS['r2'],
                         cv=self.cv)
        self.R.fit(X, Y)
        return self

    def transform(self, X):
        """Transform X using optimal transform computed during fit.
        """
        return self.R.predict(X)


class Hungarian(Alignment):
    '''Compute the optimal permutation matrix of X toward Y

    Attributes
    ----------
    R : scipy.sparse.csr_matrix
        Mixing matrix containing the optimal permutation
    '''

    def fit(self, X, Y):
        '''

        Parameters
        -----------
        X: (n_samples, n_features) nd array
            source data
        Y: (n_samples, n_features) nd array
            target data'''
        self.R = optimal_permutation(X, Y).T
        return self

    def transform(self, X):
        """Transform X using optimal permutation computed during fit.
        """
        return X.dot(self.R.toarray())


def _import_ot():
    '''Import the optional dependency ot (POT module) if installed or give
    back a clear error message to the user if not installed
    '''
    try:
        import ot
    except ImportError:
        from fmralign.version import REQUIRED_MODULE_METADATA
        for module, metadata in REQUIRED_MODULE_METADATA:
            if module == 'POT':
                POT_min_version = metadata['min_version']
        raise ImportError("To use optimal transport solver, POT module(v > {}) \
            is necessary but not installed by default with fmralign. To install \
            it run 'pip install POT'".format(POT_min_version))
    else:
        return ot


class OptimalTransportAlignment(Alignment):
    '''Compute the optimal coupling between X and Y with entropic regularization.

    Parameters
    ----------
    solver : str (optional)
        solver from POT called to find optimal coupling 'sinkhorn', \
        'greenkhorn', 'sinkhorn_stabilized','sinkhorn_epsilon_scaling', 'exact' \
        see POT/ot/bregman on Github for source code of solvers
    metric : str(optional)
        metric used to create transport cost matrix, \
        see full list in scipy.spatial.distance.cdist doc
    reg : int (optional)
        level of entropic regularization

    Attributes
    ----------
    R : scipy.sparse.csr_matrix
        Mixing matrix containing the optimal permutation
    '''

    def __init__(self, solver='sinkhorn_epsilon_scaling',
                 metric='euclidean', reg=1):
        self.ot = _import_ot()
        self.solver = solver
        self.metric = metric
        self.reg = reg

    def fit(self, X, Y):
        '''Parameters
        --------------
        X: (n_samples, n_features) nd array
            source data
        Y: (n_samples, n_features) nd array
            target data'''

        n = len(X.T)
        if n > 5000:
            warnings.warn(
                'One parcel is {} voxels. As optimal transport on this region '.format(n) +
                'would take too much time, no alignment was performed on it. ' +
                'Decrease parcel size to have intended behavior of alignment.')
            self.R = np.eye(n)
            return self
        else:
            a = np.ones(n) * 1 / n
            b = np.ones(n) * 1 / n

            M = cdist(X.T, Y.T, metric=self.metric)

            if self.solver == 'exact':
                self.R = self.ot.lp.emd(a, b, M) * n
            else:
                self.R = self.ot.sinkhorn(
                    a, b, M, self.reg, method=self.solver) * n
            return self

    def transform(self, X):
        """Transform X using optimal coupling computed during fit.
        """
        return X.dot(self.R)



from sklearn.utils.testing import assert_array_almost_equal, assert_greater
from sklearn.metrics import r2_score
from nilearn.input_data import NiftiMasker
import numpy as np
import nibabel


def assert_class_align_better_than_identity(algo, X, Y):
    """ Tests that the given algorithm align ndarrays X into Y better than \
    identity. This alignment is measured through r2 score.
    """
    print(algo)
    algo.fit(X, Y)
    identity_baseline_score = zero_mean_coefficient_determination(
        Y, X)
    algo_score = zero_mean_coefficient_determination(Y, algo.transform(X))
    assert_greater(algo_score, identity_baseline_score)


def assert_algo_transform_almost_exactly(algo, img1, img2, mask=None):
    """ Tests that the given algorithm manage to transform almost exactly Nifti\
     image img1 into Nifti Image img2
    """
    algo.fit(img1, img2)
    imtest = algo.transform(img1)
    masker = NiftiMasker(mask_img=mask)
    masker.fit()
    assert_array_almost_equal(masker.transform(
        img2), masker.transform(imtest), decimal=6)


def random_niimg(shape):
    """ Produces a random nifti image of shape (shape) and the appropriate \
    mask to use it.
    """
    im = nibabel.Nifti1Image(np.random.random_sample(shape), np.eye(4))
    mask_img = nibabel.Nifti1Image(np.ones(shape[0:3]), np.eye(4))
    return im, mask_img


def assert_model_align_better_than_identity(algo, img1, img2, mask=None):
    """ Tests that the given algorithm align Nifti image img1 into Nifti \
    Image img2 better than identity. Proficiency is measured through r2 score.
    """
    algo.fit(img1, img2)
    im_test = algo.transform(img1)
    masker = NiftiMasker(mask)
    masker.fit()
    identity_baseline_score = zero_mean_coefficient_determination(
        masker.transform(img2), masker.transform(img1))
    algo_score = zero_mean_coefficient_determination(
        masker.transform(img2), masker.transform(im_test))
    assert_greater(algo_score, identity_baseline_score)


def zero_mean_coefficient_determination(y_true, y_pred, sample_weight=None,
                                        multioutput="uniform_average"):
    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))

    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))

    if sample_weight is not None:
        weight = sample_weight[:, np.newaxis]
    else:
        weight = 1.

    numerator = (weight * (y_true - y_pred) ** 2).sum(axis=0, dtype=np.float64)
    denominator = (weight * (y_true) ** 2).sum(axis=0, dtype=np.float64)
    nonzero_denominator = denominator != 0
    nonzero_numerator = numerator != 0
    valid_score = nonzero_denominator & nonzero_numerator
    output_scores = np.ones([y_true.shape[1]])
    output_scores[valid_score] = 1 - (numerator[valid_score] /
                                      denominator[valid_score])
    output_scores[nonzero_numerator & ~nonzero_denominator] = 0

    if multioutput == 'raw_values':
        # return scores individually
        return output_scores
    elif multioutput == 'uniform_average':
        # passing None as weights results is uniform mean
        avg_weights = None
    elif multioutput == 'variance_weighted':
        avg_weights = (weight * (y_true - np.average(y_true, axis=0,
                                                     weights=sample_weight))
                       ** 2).sum(axis=0, dtype=np.float64)
        # avoid fail on constant y or one-element arrays
        if not np.any(nonzero_denominator):
            if not np.any(nonzero_numerator):
                return 1.0
            else:
                return 0.0
    else:
        avg_weights = multioutput

    return np.average(output_scores, weights=avg_weights)
